#include <v_lib.h>
#include <../source/v_mt.h>
#include "../stroke/v_stroke.h"

/*
 *  Stroke font.
 */

typedef struct stroke_font_s
{
#ifdef  CHECKED_BUILD
    u32_t   magic;
#endif  /* CHECKED_BUILD */
    u32_t   ref_count;  /* Font can be selected in several contexts */
    vec_t   height;     /* Maximal height of capital */
    vec_t   ascent;     /* Maximal distance from baseline to top */
    vec_t   descent;    /* Maximal distance from baseline to bottom */
    u32_t   num_codes;  /* Number of codes (NOT glyphs !) in font */
    vec_t   *widths;    /* Width table */
    u32_t   *code_tab;  /* Mapping from input ABC to glyph offset */
    mesh_strokes_t glyphs;  /* Glyph definitions */
} stroke_font_t, *stroke_font_p;

V_LOCAL stroke_font_p
uiConvertStrokesToFont(
    vec_t height,   /* Height of capital */
    vec_t ascent,   /* Ascent (positive) */
    vec_t descent,  /* Descent (positive or negative) */
    u32_t num_codes,
    mesh_strokes_t glyphs)
{
    u32_t extra;
    stroke_font_t *fnt;
    u32_t num_glyphs=meshGlyphCount(glyphs);
    extra=num_glyphs*sizeof(vec_t)+ /* For width table */
        num_codes*sizeof(u32_t);    /* For char mapping */
    ALLOC_OBJECT(fnt,stroke_font_t,extra,0);
    fnt->ref_count=1;
    fnt->height=height;
    fnt->ascent=ascent;
    fnt->descent=descent;
    fnt->num_codes=num_codes;
    fnt->glyphs=glyphs;
    /* Set addresses */
    fnt->widths=(vec_t*)(fnt+1);
    fnt->code_tab=(u32_t*)(fnt->widths+num_glyphs);
    /* NOTE: Caller MUST setup code table and widths ! */
    return fnt;
}
    
void    V_GLOBAL
uiDeleteFont(stroke_font_t *fnt)
{
    if(fnt->ref_count==1)
    {
        /* Delete glyphs */
        meshDeleteStrokes(fnt->glyphs);
        /* Delete object */
        memFree(fnt);
    }
    else
        fnt->ref_count--;
}

/*  =================================================================== */

/*
 *  Input:
 *  Byte 1          7   6   5   4   3   2   1   0     bit #
 *                 op1  <seven bit signed X coord>
 *  Byte 2          7   6   5   4   3   2   1   0     bit #
 *                 op2  <seven bit signed Y coord>
 *      op1=0  op2=0  End of character definition.
 *      op1=1  op2=0  Move the pointer to (x,y)
 *      op1=1  op2=1  Draw from current pointer to (x,y)
 *  Returns stroke opcode: 0 - end; 1 - move to; 2 - line to.
 */
V_LOCAL u32_t
bgiDecodeStroke(u16_t stroke,vec3_t coord)
{
    struct bgi_stroke
    {
        i16_t   x   : 7;
        u16_t   op1 : 1;
        i16_t   y   : 7;
        u16_t   op2 : 1;
    }   * ss;
    ss = ((struct bgi_stroke*)&stroke);
    vec3Set(coord,ss->x,ss->y,0);
    return ss->op2 ? 2 : ss->op1;
}

/*
 *  Reads glyph definition from BGI font.
 */
V_LOCAL vboolean
bgiReadGlyph(file_t in,mesh_strokes_t out,vec_t scale)
{
    u32_t opcode=0;
    vec3_t coord, origin, pt;
    while(!fileEof(in))
    {
        opcode=bgiDecodeStroke(fileGetLittleShort(in),coord);
        if(opcode==2)
        {
            /* LineTo: store coord */
            vec3Scale(pt,coord,scale);
            meshStrokeCoordv(out,pt);
        }
        else
        {
            /* End of MoveTo: begin new stroke */
            meshEndStroke(out);
            if(!opcode) /* End */
                break;
            vec2Copy(origin,coord); /* MoveTo */
            vec3Scale(pt,coord,scale);
            meshStrokeCoordv(out,pt);
        }
    }
    if(opcode)
    {
        /* Error: unexpected EOF; discard all coords and strokes */
        meshCancelStroke(out);
        meshCancelGlyph(out);
    }
    else
    {
        /* Create glyph */
        meshEndGlyph(out);
    }
    return opcode==0;
}

V_LOCAL u32_t bgi_koi2alt[256]=
{
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
    0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
    0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
    0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
    0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
    0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
    0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
    0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
    0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
    0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
    0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
    0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
    0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
    0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
    0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
    0xc4, 0xb3, 0xda, 0xbf, 0xc0, 0xd9, 0xc3, 0xb4,
    0xc2, 0xc1, 0xc5, 0xdf, 0xdc, 0xdb, 0xdd, 0xde,
    0xb0, 0xb1, 0xb2, 0xf4, 0xfe, 0xf9, 0xfb, 0xf7,
    0xf3, 0xf2, 0xff, 0xf5, 0xf8, 0xfd, 0xfa, 0xf6,
    0xcd, 0xba, 0xd5, 0xf1, 0xd6, 0xc9, 0xb8, 0xb7,
    0xbb, 0xd4, 0xd3, 0xc8, 0xbe, 0xbd, 0xbc, 0xc6,
    0xc7, 0xcc, 0xb5, 0xf0, 0xb6, 0xb9, 0xd1, 0xd2,
    0xcb, 0xcf, 0xd0, 0xca, 0xd8, 0xd7, 0xce, 0xfc,
    0xee, 0xa0, 0xa1, 0xe6, 0xa4, 0xa5, 0xe4, 0xa3,
    0xe5, 0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae,
    0xaf, 0xef, 0xe0, 0xe1, 0xe2, 0xe3, 0xa6, 0xa2,
    0xec, 0xeb, 0xa7, 0xe8, 0xed, 0xe9, 0xe7, 0xea,
    0x9e, 0x80, 0x81, 0x96, 0x84, 0x85, 0x94, 0x83,
    0x95, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e,
    0x8f, 0x9f, 0x90, 0x91, 0x92, 0x93, 0x86, 0x82,
    0x9c, 0x9b, 0x87, 0x98, 0x9d, 0x99, 0x97, 0x9a
};

V_LOCAL u32_t bgi_win2alt[256]=
{
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
    0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
    0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
    0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
    0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
    0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
    0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
    0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
    0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
    0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
    0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
    0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
    0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
    0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
    0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
    0x3f, 0x3f, 0x27, 0x3f, 0x22, 0x3a, 0xc5, 0xd8,
    0x3f, 0x25, 0x3f, 0x3c, 0x3f, 0x3f, 0x3f, 0x3f,
    0x3f, 0x27, 0x27, 0x22, 0x22, 0x07, 0x2d, 0x2d,
    0x3f, 0x54, 0x3f, 0x3e, 0x3f, 0x3f, 0x3f, 0x3f,
    0xff, 0xf6, 0xf7, 0x3f, 0xfd, 0x3f, 0xb3, 0x15,
    0xf0, 0x63, 0xf2, 0x3c, 0xbf, 0x2d, 0x52, 0xf4,
    0xf8, 0x2b, 0x49, 0x69, 0x3f, 0xe7, 0x14, 0xfa,
    0xf1, 0xfc, 0xf3, 0x3e, 0x3f, 0x3f, 0x3f, 0xf5,
    0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
    0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
    0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,
    0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
    0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
    0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
    0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,
    0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef
};

stroke_font_p V_GLOBAL
uiLoadBGIFont(char *file_name,vboolean load_as_koi8)
{
    file_t in;
    u32_t glyphs_base;
    u32_t *offsets;
    u32_t i,numchars,firstchar;
    u32_t *recode;
    i32_t top,baseline,bottom;
    u8_t *widths;
    vec_t ascent, descent, height;
    vec_t scale;
    mesh_strokes_t out;
    stroke_font_t *fnt;
    vboolean success;
    in=fileOpenRead(file_name);
    if(!in)
        return NULL;
    if(fileGetLittleLong(in)!=0x08084B50)
    {
#ifdef  CHECKED_BUILD
        conlPrintf(LOG_ERROR,"uiLoadBGIFont(%s): isn't BGI font.\n",
            file_name);
#endif  /* CHECKED_BUILD */
        fileClose(in);
        return NULL;
    }
    fileSeek(in,0x81,SEEK_SET);
    /* Read header */
    numchars=fileGetLittleShort(in);
    fileGetc(in);
    firstchar=fileGetc(in);
    fileSeek(in,3,SEEK_CUR);
    top=fileGetc(in);
    baseline=fileGetc(in);
    bottom=fileGetc(in);
    ascent=(vec_t)top-(vec_t)baseline;
    descent=(vec_t)bottom-(vec_t)baseline;
    height=(vec_t)ascent-(vec_t)descent;
    scale=1/height;
    height=1;
    ascent*=scale;
    descent*=scale;
    /* Read offset table */
    fileSeek(in,0x90,SEEK_SET);
    offsets=(u32_t*)memAlloc(numchars*sizeof(u32_t));
    for(i=0;i<numchars;i++)
        offsets[i]=fileGetLittleShort(in);
    /* Read width table */
    widths=(u8_t*)memAlloc(numchars);
    fileRead(in,widths,numchars);
    /* Read strokes */
    success=true;
    out=meshCreateStrokes();
    glyphs_base=fileTell(in);
    for(i=0;i<numchars;i++)
    {
        fileSeek(in,glyphs_base+offsets[i],SEEK_SET);
        success = success && bgiReadGlyph(in,out,scale);
    } 
    memFree(offsets);
    if(success)
        fnt=uiConvertStrokesToFont(height,ascent,descent,256,out);
    else
        fnt=NULL;
    if(fnt)
    {
        /* Setup width table */
        for(i=0;i<numchars;i++)
            fnt->widths[i]=scale * (vec_t)widths[i];
        /* Setup code table */
        if(load_as_koi8)
            recode=bgi_koi2alt;
        else
            recode=bgi_win2alt;
        memZero(fnt->code_tab,256*sizeof(u32_t));
        for(i=0;i<256;i++)
        {
            if(i<firstchar)
                fnt->code_tab[i]=firstchar;
            else
                fnt->code_tab[i]=recode[i]-firstchar;
        }
    }
    else
    {
#ifdef  CHECKED_BUILD
        conlPuts(LOG_ERROR,
            "uiLoadBGIFont(): unable to read all glyphs.\n");
#endif  /* CHECKED_BUILD */
        /* Delete strokes object */
        meshDeleteStrokes(out);
    }
    memFree(widths);
    fileClose(in);
    return fnt;
}

void    V_GLOBAL
uiDrawChar(stroke_font_t *font,u32_t ch,vec_t *dx_return)
{
    vec_t dx;
    u32_t glyph,i,num_strokes,j,num_coords;
    mesh_stroke_t *stroke;
    if(ch>=font->num_codes)
        if('?'<font->num_codes)
            ch='?';
        else
            return;
    glyph=font->code_tab[ch];
    num_strokes=meshStrokesCount(font->glyphs,glyph);
    dx=font->widths[glyph];
    for(i=0;i<num_strokes;i++)
    {
        stroke=meshGlyphStroke(font->glyphs,glyph,i);
        num_coords=stroke->num_coords;
        glBegin(GL_LINE_STRIP);
        for(j=0;j<num_coords;j++)
            GL_VERTEX3V(stroke->coords[j]);
        glEnd();
    }
    glTranslatef(dx,0,0);
    if(dx_return)
        *dx_return=dx;
}

void    V_GLOBAL
uiDrawStringEx(stroke_font_t *font,void *str,u32_t charsize)
{
    vec_t cw, dx, sw, ts;
    u32_t ch;
    u8_t *ptr;
    if(charsize!=1 && charsize!=2 && charsize!=4)
    {
#ifdef  CHECKED_BUILD
        conlPrintf(LOG_ERROR,"uiDrawStringEx(): wrong char size '%u'.\n",
            charsize);
#endif  /* CHECKED_BUILD */
        return;
    }
    sw=font->widths[font->code_tab[0x20]]; /* space width */
    dx=0;       /* accumulate increment to avoid GL stack pushes/pops */
    ts=sw*4;    /* TODO: implement real tabs alignment */
    for(ptr=(unsigned char*)str;*ptr;ptr+=charsize)
    {
        switch(charsize)
        {
            case 4:
                ch=*(u32_t*)ptr;
                break;
            case 2: ch=*(u16_t*)ptr;
                ch=*(u16_t*)ptr;
                break;
            default:
                ch=*ptr;
        }
        switch(ch)
        {
            case '\t':
                glTranslatef(ts,0,0);
                dx+=ts;
                break;
            case ' ':
                glTranslatef(sw,0,0);
                dx+=sw;
                break;
            case 0x0D:
                break;
            case 0x0A:
                glTranslatef(-dx,-font->height,0);
                dx=0;
                break;
            default:
                uiDrawChar(font,ch,&cw);
                dx+=cw;
        }
    }
    return;
}

void    V_GLOBAL
uiDrawStringA(stroke_font_t *font,char *str)
{
    uiDrawStringEx(font,str,1);
}

void    V_GLOBAL
uiDrawStringW(stroke_font_t *font,char *str)
{
    uiDrawStringEx(font,str,2);
}

void    V_GLOBAL
uiDrawKoi8r(stroke_font_t *font,char *str)
{
    static char koi8abc[]="JCUKENG[]ZH_FYWAPROLDV\\"
        "Q^SMITXB@#jcukeng{}zhfywaproldv|q~smitxb`3";
    u8_t *ptr,*koi=(u8_t*)strDup(str);
    for(ptr=koi;*ptr;ptr++)
    {
        if(strchr(koi8abc,*ptr))
            *ptr+=128;
    }
    uiDrawStringEx(font,koi,1);
    memFree(koi);
}

void V_CALL draw_scene(void *param)
{
    stroke_font_t *f=(stroke_font_t*)param;
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-1,1,-1,1,-1,1);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glDisable(GL_DEPTH_TEST);
    glTranslatef(-1,1,0);
    glScalef(.3f,.3f,.3f);
    glTranslatef(0,-1,0);
    glClearColor(0,0,0,0);
    glClear(GL_COLOR_BUFFER_BIT);
    glColor4f(1,1,1,.8f);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    glEnable(GL_LINE_SMOOTH);
    glLineWidth(2.5f);
    uiDrawKoi8r(f,"pRIWET !\n");
    uiDrawKoi8r(f,"\n1234567890");
}

int main()
{
    stroke_font_t *f;
    logStartConsoleLog();
    logUpperLevel(LOG_TRACE);
    f=uiLoadBGIFont("../../samples/media/bgi-fonts/cyrgoth.chr",true);
    if(f)
    {
        input_record_t r;
        vglCreateFullscreenWindow("",0,0,0,0);
        draw_scene(f);
        vglSwapBuffers();
        do
            vglNextEvent(&r);
        while(r.event!=ev_key_down);
        /* vglDrawToMetafile("//./clipboard",100,100,draw_scene,f); */
        vglDestroyWindow(vglGetCurrentWindow());
        uiDeleteFont(f);
    }
    return 0;
}
